@@ -131,12 +131,17 @@ static void mp_msg_av_log_callback(void *ptr, int level, const char *fmt,
     struct mp_log *log = get_av_log(ptr);
 
     if (mp_msg_test(log, mp_level)) {
+        char buffer[4096] = "";
+        int pos = 0;
         const char *prefix = avc ? avc->item_name(ptr) : NULL;
         if (log_print_prefix && prefix)
-            mp_msg(log, mp_level, "%s: ", prefix);
+            pos = snprintf(buffer, sizeof(buffer), "%s: ", prefix);
         log_print_prefix = fmt[strlen(fmt) - 1] == '\n';
 
-        mp_msg_va(log, mp_level, fmt, vl);
+        pos = MPMIN(MPMAX(pos, 0), sizeof(buffer));
+        vsnprintf(buffer + pos, sizeof(buffer) - pos, fmt, vl);
+
+        mp_msg(log, mp_level, "%s", buffer);
     }
 
     pthread_mutex_unlock(&log_lock);
@@ -202,17 +207,29 @@ void print_libav_versions(struct mp_log *log, int v)
 
     mp_msg(log, v, "%s library versions:\n", LIB_PREFIX);
 
+    bool mismatch = false;
     for (int n = 0; n < MP_ARRAY_SIZE(libs); n++) {
         const struct lib *l = &libs[n];
         mp_msg(log, v, "   %-15s %d.%d.%d", l->name, V(l->buildv));
-        if (l->buildv != l->runv)
+        if (l->buildv != l->runv) {
             mp_msg(log, v, " (runtime %d.%d.%d)", V(l->runv));
+            mismatch = true;
+        }
         mp_msg(log, v, "\n");
     }
 
 #if HAVE_AV_VERSION_INFO
     mp_msg(log, v, "%s version: %s\n", LIB_PREFIX, av_version_info());
 #endif
+
+    if (mismatch) {
+        // Using mismatched libraries can be legitimate, but even then it's
+        // a bad idea. We don't acknowledge its usefulness and stability.
+        mp_warn(log, "Warning: mpv was compiled against a different version of "
+                "%s than the shared\nlibrary it is linked against. This is "
+                "most likely a broken build\nand misbehavior and crashes are "
+                "to be expected.\n", LIB_PREFIX);
+    }
 }
 
 #undef V
