@@ -23,7 +23,8 @@
 #include <stdint.h>
 #include <stdbool.h>
 
-#include "config.h"
+#include <ass/ass.h>
+#include <ass/ass_types.h>
 
 // This is probably arbitrary.
 // sd_lavc_conv might indirectly still assume this PlayResY, though.
@@ -35,10 +36,6 @@
 // m_color argument
 #define MP_ASS_COLOR(c) MP_ASS_RGBA((c).r, (c).g, (c).b, (c).a)
 
-#if HAVE_LIBASS
-#include <ass/ass.h>
-#include <ass/ass_types.h>
-
 struct MPOpts;
 struct mpv_global;
 struct mp_osd_res;
@@ -55,7 +52,6 @@ ASS_Library *mp_ass_init(struct mpv_global *global, struct mp_log *log);
 struct sub_bitmap;
 struct sub_bitmaps;
 void mp_ass_render_frame(ASS_Renderer *renderer, ASS_Track *track, double time,
-                         struct sub_bitmap **parts, struct sub_bitmaps *res);
+                         struct sub_bitmaps *res);
 
-#endif                          /* HAVE_LIBASS */
 #endif                          /* MPLAYER_ASS_MP_H */
